{Start of file PrimeFA.PAS *************************************************}

{$A+,B-,D+,E+,F-,G-,I+,L+,N-,O-,R-,S+,V+,X-}
{$M 16384,0,655360}

{$I-} {do our own i/o error checks}
{$N+} {Uses numeric coprocessor}
{$R-} {index Range checking}
{$V-} {no string length checking}
{$M 65520, 0, 655360} {16384, 0, 655360 are defaults; $M 65520 max stack}

program PrimeFA; {Multi-precision prime factors algorithm}

uses
  Dos, Crt, { Turbo Pascal interface }
  MultiID;  { Multiple-precision integer decimal algorithms unit }

const
  Name    = 'PrimeFA - Multi-precision prime factor algorithm';
  Version = 'Version 2.50, last revised: 1994-11-20, 1100 hours';
  Author  = 'Copyright (c) 1981-1994 by author: Harry J Smith,';
  Address = '19628 Via Monte Dr., Saratoga CA 95070.  All rights reserved.';

{***************************************************************************}

{Developed in Turbo Pascal 5.0 and converted to 6.0}
{Factoring into primes.  See Knuth, Vol. 2, Page 348}

const
  Size = 1429; {29 for factors of 101 decimal digit numbers if MuDMax = 7}
               {Max number of "Digits" in a multi-precision number}
               {Each "Digit" is from 0 to 999...9, i.e. Base = 10 ** MuDMax}
  Good = True;

const
  LenPTab = 3512; { 200 normally, 3512 at max }
  RS      = 8;    { Number of sieve tables (r) for SieveMethod }
  MRM     = 22;   { MRM= PTab[RS + 1] - 1, (m sub r - 1) for SieveMethod }
  MaxPs   = 1400; { Max number of prime factors, 24 for 36 digit numbers }
                  { 54 for 101 digit numbers, about 1400 for 5000 digit #s }

TYPE
  TPrimes   = Array[1..MaxPs] of ^MultiSI; {Prime factor table}
  TPowers   = Array[1..MaxPs] of Integer;  {Factor power table}

var         {Global variables}
  ADiagOn   : Boolean; { Application Diagnostics turned on }
  FindAPrime: Boolean; {Find first prime < ZIn or > ZIn}
  FindUp    : Boolean; {Find first prime > ZIn, See I toggle}
  PTabFilled: Boolean; { PTab has been filled }
  NoSieve   : Boolean; { Do not use sieve method }
  NoMoreDiv : Boolean; { Do not use MoreDiv method }
  ZIn       : ^MultiSI;{Original input number to be factored}
  Start     : ^MultiSI;{Minimum factor to start prime test, normally 2}
  CutOff    : ^MultiSI;{ Limit for using MethodX }
  EI        : Integer; {Index for Echo to printer}
  D         : Array[-1..10] of Integer; {Constant table used in MoreDiv}
  PTab      : Array[1..LenPTab] of Integer; {Small prime table}
  SqrMaxP   : Double;  {Square of max prime in PTab}
  Sieve     : Array[1..RS, 0..MRM] of Boolean; {For SieveMethod} {RS tables}
  SqrMod    : Array[0..MRM] of Integer;        { "     " }
  KS        : Array[1..RS] of Integer;         { "     " }
  SieveN    : ^MultiSI;                        { "     " }
  RemS      : ^MultiSI;                        { "     " }
  Root2     : ^MultiSI;                        { "     " }
  DelX      : ^MultiSI;                        { "     " }
  StIn      : String[128]; { Operator input string }
  FileSt    : String[12];  { Output file name }
  Disk      : Text;        { Text file }

{--------------------------------------}
procedure ADiag(Mess : String); {Output a diagnostic message with delta times}
var
  EI   : Integer;
  Del0 : Double; {Total time from TV0}
  Del1 : Double; {Delta time from last Diag call}
begin
  if ADiagOn then begin
    TV1:= TV2;  TV2:= DosClock;
    Del0:= TV2 - TV0;
    Del1:= TV2 - TV1;
    for EI:= 0 to Echo do
      WriteLn(OutP[EI]^, 'T = ', Del0:0:2,'  DT = ',Del1:0:2,' sec.  ',Mess);
  end;
end; { ADiag }

{--------------------------------------}
procedure FillPTab; {Fill the prime table PTab[1..LenPTab]}
var
  N     : Integer; {Number to test if it is prime 3, 5, 7, ...}
  I     : Integer; {Index to old entry in prime table}
  J     : Integer; {Index to new entry in prime table}
  Di    : Integer; {Divisor to test, an old entry in prime table}
  Q     : Integer; {Quotient, N / Di}
  RZero : Boolean; {True if Remainder is zero}
begin
  PTab[1]:= 2;  PTab[2]:= 3;
  N:= 3;  J:= 2;
  repeat
    N:= N + 2;  I:= 2;
    repeat
      Di:= PTab[I];  I:= I + 1;
      Q:= N div Di;
      RZero:= (N - Q * Di) = 0;
    until (Di >= Q) or RZero;
    if Not RZero then begin
      J:= J + 1;  PTab[J]:= N;
    end;
  until J = LenPTab;
  SqrMaxP:= PTab[LenPTab];
  SqrMaxP:= SqrMaxP * SqrMaxP;
end; {FillPTab}

{--------------------------------------}
procedure Init; { Initialize program }
var  i : Integer;
begin
  TextBackground( Blue);
  TextColor( Yellow);
  ClrScr;
  D[-1]:= 0;
  D[0]:= 1;  D[1]:= 2;  D[2]:= 2;
  D[3]:= 4;  D[4]:= 2;  D[5]:= 4;
  D[6]:= 2;  D[7]:= 4;  D[8]:= 6;
  D[9]:= 2;  D[10]:= 6;
  ADiagOn:= False;  FindAPrime:= False;
  ADiag('Start PrimeFA');
  PTabFilled:= False;
  NoSieve   := False;
  NoMoreDiv := False;
  FindUp    := False;
  MultiIDV.Init( Size);  { Init Multiple-precision integer decimal algorithms }
  MuLenD:= 21;  MuErrRep:= True;  DiagOn:= True;
  New( ZIn,    Init( Size));        { Create new objects }
  New( Start,  Init( Size));
  New( SieveN, Init( Size));
  New( RemS,   Init( Size));
  New( Root2,  Init( Size));
  New( DelX,   Init( Size));
  New( CutOff, Init( 3));
  CutOff^.Value( '18446,74407,37095,51615', i);
  FileSt:= 'PrimeFA.Out';
end; { Init }

{--------------------------------------}
procedure Wr2Ln( St : String); { WriteLn to both the Screen and Disk }
begin
  for EI:= 0 to Echo do  WriteLn( OutP[EI]^, St);
end; { Wr2Ln }

{--------------------------------------}
procedure HelpO(var Out : TEXT);{Give basic help information to the operator}
var
  I : Integer;
  St1 : String[6];
  St2 : String[10];
begin
  St1:= '';  St2:= '';  I:= Test8087;
  if I = 0 then  St2:= ' emulation';
  if I = 1 then  St1:= ' 8087';
  if I = 2 then  St1:= ' 80287';
  if I = 3 then  St1:= ' 80387';
  WriteLn( Out);
  WriteLn( Out);
  WriteLn( Out, Name);
  WriteLn( Out, Version);
  WriteLn( Out, Author);
  WriteLn( Out, Address);
  WriteLn( Out);
  WriteLn( Out, 'Prime factors of Integers <= ', MuDMax * Size div 2,
    ' decimal digits (using', St1, ' coprocessor', St2, ')');
  WriteLn( Out, 'Some may take forever (over 300,000,000 years)');
  WriteLn( Out);
  WriteLn( Out, 'Enter Q to Quit, E to Echo screen to file: ', FileSt, ',');
  Write  ( Out, 'D for Diagnostics, S to force Sieve method,');
  WriteLn( Out, ' T to toggle find direction,');
  WriteLn( Out, 'M to force More divide method, N for Normal');
  WriteLn( Out);
  WriteLn( Out, 'Can also enter numbers like 5*(2+3)/2-3^(2!)');
  WriteLn( Out, '(= => +,   . => *,   , => ignored,   P => ^,   F => !,');
  WriteLn( Out, '? => Help,   -number => find a prime)');
  WriteLn( Out);
  WriteLn( Out, 'Or enter numbers like 1234s12');
  WriteLn( Out, 'to start testing 1234 at start divisor 12');
  WriteLn( Out);
  WriteLn( Out, 'Or enter numbers like 9r8,');
  WriteLn( Out, 'to repeat 9, 8 times i.e. 99999999');
  WriteLn( Out, '(use ESC to interrupt if something runs too long)');
end; {HelpO}

{--------------------------------------}
procedure HelpE; { Give basic help information to the operator with echo }
begin
  if Echo = 1 then begin
    HelpO( OutP[1]^);  WriteLn( OutP[1]^);
  end;
  ClrScr;
  HelpO( OutP[0]^);
end; { HelpE }

{--------------------------------------}
procedure Fini;
begin
  Dispose( ZIn,    Done); { Clean up your act }
  Dispose( Start,  Done);
  Dispose( SieveN, Done);
  Dispose( RemS,   Done);
  Dispose( Root2,  Done);
  Dispose( DelX,   Done);
  Dispose( CutOff, Done);
  Close( Disk);
  Halt(0);
end; { Fini }

{--------------------------------------}
procedure TEcho; { Toggle Echo to file }
begin
  Echo:= (Echo + 1) Mod 2;
  if Echo = 1 then begin
    Assign( Disk, FileSt);
    Append( Disk);  OutP[1]:= @Disk;
    if IOResult <> 0 then begin
      Rewrite( Disk);  OutP[1]:= @Disk;
      Wr2Ln('Will create and write file: ' + FileSt);
    end else
      Wr2Ln('Will append to file: ' + FileSt);
  end
  else begin
    Wr2Ln( 'Closed file: ' + FileSt);
    Close( Disk);
  end;
end; { TEcho }

{--------------------------------------}
procedure PowerMod(B : Mu1Digit; var E, M, PM : MultiSI); {PM:= (B ^ E) mod M}
var
  BL : ^MultiSI; {Local value of B}
begin
  if MuAbort then Exit;
  if ADiagOn then  ADiag('Enter PowerMod');
  New( BL, Init( Size));
  BL^.SetTo1( B);
  if KeyHit or ADiagOn then begin
    KeyHit:= False;
    for EI:= 0 to Echo do begin
      Write( OutP[EI]^, B:0:0);  Write( OutP[EI]^, ' ^ ');  E.Writ( OutP[EI]^);
      Write( OutP[EI]^, ' mod ');
      M.Writ( OutP[EI]^);  Write( OutP[EI]^, ' = ');
    end;
  end;
  MuMB^.SetTo( M);
  PM.PowMB(BL^, E);
  MuMB^.Clear;
  if KeyHit or ADiagOn then begin
    KeyHit:= False;
    for EI:= 0 to Echo do  PM.WritLn( OutP[EI]^);
  end;
  Dispose( BL, Done);
  if ADiagOn then  ADiag('Exit PowerMod');
end; {PowerMod}

{--------------------------------------}
procedure DelPrimes(var DNF     : Integer;  { Delete prime factors found }
                    var DPrimes : TPrimes);
var  I : Integer;
begin
  for I:= 1 to DNF do
    Dispose( DPrimes[I], Done);
  DNF:= 0;
end; { DelPrimes }

{--------------------------------------}
procedure FactorN(var N      : MultiSI;  {Factor N into its prime factors}
	          var Start  : MultiSI;  {Minimum factor to start prime test}
		      Report : Boolean;  {True to report factors to operator}
                  var FNF    : Integer;  {Number of prime factors found}
                  var FPrimes: TPrimes;  {Prime factors found}
                  var FPowers: TPowers); {Powers of the prime factors found}

var Found      : Boolean; {Factor found}
    TwoFactors : Boolean; {Only two prime factors in number}
    ExitFactor : Boolean; {Exit Factor flag}
    Di         : ^MultiSI;{Divisor to test}
    Po          : Integer; {Power of prime factor}

{--------------------------------------}
procedure ReportFactor; {Part of FactorN, reports Di as a factor}
begin
  if MuAbort then Exit;
  if ADiagOn then  ADiag('Enter ReportFactor');
  ExitFactor:= False;
  if KeyHit or ADiagOn then begin
    KeyHit:= False;
    for EI:= 0 to Echo do begin
      Write( OutP[EI]^, 'Report=');
      if Report then  Write( OutP[EI]^, 'True')
                else  Write( OutP[EI]^, 'False');
      Write( OutP[EI]^, ' NF=', FNF + 1, ' Po=', Po, ' Factor=');
      Di^.WritLn( OutP[EI]^);
    end;
  end;
  if FNF < MaxPs then begin
    FNF:= FNF + 1;
    New( FPrimes[FNF], Init( Di^.N));
    FPrimes[FNF]^.SetTo(Di^);
    FPowers[FNF]:= Po;
  end;
  if Report then begin
    for EI:= 0 to Echo do begin
      Di^.Writ( OutP[EI]^);  WriteLn( OutP[EI]^, ' ^ ', Po);
    end;
  end
  else begin
    if ADiagOn then  ADiag('Exit FactorN from ReportFactor');
    ExitFactor:= True;
    Exit;
  end;
  if ADiagOn then  ADiag('Exit ReportFactor');
end; {ReportFactor}

{--------------------------------------}
procedure SmallFactors; {Part of FactorN, find small factors in PTab}
var
  I     : Integer; {Index}
  Si    : Integer; {Sign result of AbComp procedure}
  Q     : ^MultiSI;{Quotient}
  R     : ^MultiSI;{Remainder of divide}
  RZero : Boolean; {True if Remainder is zero}
begin
  if ADiagOn then  ADiag('Enter SmallFactors');
  New( Q, Init( Size));
  New( R, Init( Size));
  ExitFactor:= False;
  I:= 1;
  repeat
    Di^.SetTo1(PTab[I]);  I:= I + 1;  Po:= 0;
    repeat
      Q^.DivQR(N, Di^, R^); {Q = N div Di}
      RZero:= R^.ZTest;
      if RZero then begin
        N.SetTo(Q^);  Po:= Po + 1;
      end;
    until Not RZero or MuAbort;
    if Po > 0 then begin
      ReportFactor;
      if (Report and FindAPrime) or ExitFactor then begin
        Dispose( Q, Done);
        Dispose( R, Done);
        if ADiagOn then  ADiag('Quick exit SmallFactors');
        Exit;
      end;
    end;
    Si:= Di^.AbComp(Q^);
  until (Si >= 0) or (I > LenPTab) or MuAbort;
  if Not N.EQ1(1.0) and Not N.GT1(SqrMaxP) then begin
    Di^.SetTo(N);  Po:= 1;  N.SetTo1(1.0);  ReportFactor;
    if ExitFactor then begin
      Dispose( Q, Done);
      Dispose( R, Done);
      if ADiagOn then  ADiag('ReportFactor exit SmallFactors');
      Exit;
    end;
  end;
  Dispose( Q, Done);
  Dispose( R, Done);
  if ADiagOn then  ADiag('Exit SmallFactors');
end; {SmallFactors}

{--------------------------------------}
procedure MoreDiv; {Part of FactorN, try the division method some more}
var
  I     : Integer; {Index}
  J     : Integer; {Index}
  Si    : Integer; {Sign result of AbComp procedure}
  Q     : ^MultiSI;{Quotient}
  R     : ^MultiSI;{Remainder of divide}
  MaxDi : ^MultiSI;{Max divisor this try of MoreDiv}
  Done  : Boolean; {Done flag}
  RZero : Boolean; {True if Remainder is zero}
  T     : Double;  {Value of DosClock}

{--------------------------------------}
procedure ExitMoreDiv; {Part of MoreDiv}
begin
  Dispose( Q, Done);
  Dispose( R, Done);
  Dispose( MaxDi, Done);
  T:= DosClock;
end; {ExitMoreDiv}

{--------------------------------------}
procedure MethodX; { Extended floating point Method, part of MoreDiv }
var
  Dx, Qx, Nx, Rx : Extended;
  J              : Integer;
begin
  J:= 0;
  Di^.GetEx( Dx);  N.GetEx( Nx);
  repeat
    repeat
      Qx:= Nx / Dx;
{     RZero:= (Frac( Qx) = 0.0); . This failed in coprocessor emulation mode}
      Rx:= Nx - Dx * Int( Qx);         { Replaced with this }
      RZero:= (Rx = 0.0) or (Rx = Dx); { Assume Qx can be off +/- epsilon }
      if RZero then begin
        Nx:= Qx;  Po:= Po + 1;  Done:= True;
      end;
      Inc( J);
      if J = 1000 then begin
        J:= 0;
        if KeyPressed then  MuInterrupt;
        if KeyHit then begin
          KeyHit:= False;
          N.SetToEx( Nx);  Di^.SetToEx( Dx);
          N.ShortWr( OutP[0]^, 40);  Write( OutP[0]^, ' / ');
          Di^.Writ( OutP[0]^);  WriteLn( OutP[0]^, '  MethodX');
        end;
      end;
    until Not RZero or MuAbort;
    if not Done then begin
      Dx:= Dx + D[I];  { Di:= Di + D[I]; }
      if I < 10 then  I:= I + 1
                else  I:= 3;
    end;
  until (Dx >= Qx) or MuAbort or Done;
  Q^.SetToEx( Qx);  N.SetToEx( Nx);  Di^.SetToEx( Dx);
end; { MethodX }

{--------------------------------------}
begin {MoreDiv}
  if NoMoreDiv then Exit;
  if ADiagOn then  ADiag('Enter MoreDiv');
  for EI:= 0 to Echo do begin
    WriteLn( OutP[EI]^);
    WriteLn( OutP[EI]^, 'Try the division method some more');
  end;
  ExitFactor:= False;
  New( Q, Init( Size));
  New( R, Init( Size));
  New( MaxDi, Init( Size));
  Found:= False;
  Si:= Start.AbComp(N);
  if Si < 0 then begin
    MaxDi^.Add1(Start, 1000000.0);
    Done:= False;  J:= 0;  Po:= 0;
    Di^.Add1(Start, -7.0);    {Di = (Start - 7) div 1020 * 1020 + 7}
    Di^.Div1(Di^, 1020.0);
    Di^.Mul1(Di^, 1020.0);
    Di^.Add1(Di^, 7.0);
    I:= 3;
    repeat
      Di^.Add1(Di^, D[I]);  {Di = Di + D[I]}
      if I < 10
      then I:= I + 1
      else begin
        I:= 3;  J:= J + 1;
        if J = 34 then begin
          J:= 0;
          T:= DosClock;
          if TwoFactors then begin
            Si:= Di^.AbComp(MaxDi^);
            if (Si > 0) then
              Done:= True
          end
          else begin
            Q^.Divi(Q^, Di^);  Si:= Di^.AbComp(Q^); { Stop at Cube Root }
            if (Si > 0) then
              Done:= True;
          end;
          if KeyHit or ADiagOn then begin
            KeyHit:= False;
            N.ShortWr( OutP[0]^, 40);  Write( OutP[0]^, ' / ');
            Di^.Writ( OutP[0]^);  WriteLn( OutP[0]^, '  MoreDiv');
            if Done and (Echo > 0) then begin
              N.ShortWr( OutP[1]^, 40);  Write( OutP[1]^, ' / ');
              Di^.Writ( OutP[1]^);  WriteLn( OutP[1]^, '  MoreDiv');
            end;
          end;
          if MuAbort then begin
            Done:= True;
            ExitFactor:= True;
            ExitMoreDiv;
            Exit;
          end;
        end;
      end;
      if (N.Comp(CutOff^) <= 0) then  MethodX
      else begin
        repeat
          Q^.DivQR(N, Di^, R^);
          RZero:= R^.ZTest;
          if RZero then begin
            N.SetTo(Q^);  Po:= Po + 1;  Done:= True;
          end;
        until Not RZero or N.ZTest;
      end;
    until Done;
    Found:= (Po > 0);
    if Found then begin
      ReportFactor;
      if ExitFactor then begin
        ExitMoreDiv;
        if ADiagOn then  ADiag('ReportFactor exit MoreDiv');
        Exit;
      end;
    end;
    Si:= Di^.AbComp(Q^);
    if Not N.EQ1(1.0) and (Si >= 0) then begin
      Di^.SetTo(N);  Po:= 1;  N.SetTo1(1.0);  ReportFactor;
      if ExitFactor then begin
        ExitMoreDiv;
        if ADiagOn then  ADiag('ReportFactor exit MoreDiv');
        Exit;
      end;
    end;
    Start.Add1(Di^, 2.0);
    Q^.Divi(Q^, Di^);  Si:= Di^.AbComp(Q^);
    TwoFactors:= (Si > 0); {(Di > Q div Di), Di > cube root of N}
  end;
  ExitMoreDiv;
  if ADiagOn then  ADiag('Exit MoreDiv');
end; {MoreDiv}

{--------------------------------------}
procedure LargeFactor; {Part of FactorN, determine if large factor is prime}
var
  NM      : ^MultiSI;{N - 1}
  NP      : ^MultiSI;{N', a secondary number to be factored}
  M       : ^MultiSI;{Result of i ^ (N-1) mod N calculation}
  Q       : ^MultiSI;{Quotient}
  PM      : ^MultiSI;{Result of PowerMod calculation for secondary number}
  Start   : ^MultiSI;{Minimum factor to start prime test, = 2}
  J       : Integer; {Index in PTab}
  X       : Integer; {Base of PowerMod calculation for prime test}
  MaxJ    : Integer; {The number of primes that have been tested with PowerMod}
  LNF     : Integer; {Number of prime factors found in NP}
  LPrimes : TPrimes; {Prime factors found in NP}
  LPowers : TPowers; {Powers of the prime factors found in NP}
begin
  if ADiagOn then  ADiag('Enter LargeFactor');
  ExitFactor:= False;
  Found:= False;
  if MuAbort then begin
    ExitFactor:= True;  Exit;
  end;
  LNF:= 0;
  New( NM,    Init( Size));
  New( NP,    Init( Size));
  New( M,     Init( Size));
  New( Q,     Init( Size));
  New( PM,    Init( Size));
  New( Start, Init( Size));
  NM^.Add1(N, -1.0);  {NM = N - 1}
  PowerMod(3.0, NM^, N, M^);  {M = (3 ^ NM) mod N}
  if Not M^.GT1(1.0) then begin
    PowerMod(2.0, NM^, N, M^);
    if Not M^.GT1(1.0) then begin
      PowerMod(5.0, NM^, N, M^);
      if Not M^.GT1(1.0) then begin
        PowerMod(7.0, NM^, N, M^);
        MaxJ:= 4;  NP^.SetTo(NM^);  Start^.SetTo1(2.0);
        while Not NP^.EQ1(1.0) and Not M^.GT1(1.0) do begin
                     {(NP <> 1) and (M < 2)}
          DelPrimes(LNF,  LPrimes);
          FactorN(NP^, Start^, False, LNF, LPrimes, LPowers);
          if MuAbort then begin
            ExitFactor:= True;
            DelPrimes(LNF,  LPrimes);
            Dispose( NM,    Done);
            Dispose( NP,    Done);
            Dispose( M,     Done);
            Dispose( Q,     Done);
            Dispose( PM,    Done);
            Dispose( Start, Done);
            Exit;
          end;
          J:= 1;
          while J <= LenPTab do begin
            X:= PTab[J];
            if J > MaxJ then begin
              PowerMod(X, NM^, N, M^);
              MaxJ:= J;
            end;
            if Not M^.GT1(1.0) then begin  {M < 2}
              Q^.Divi(NM^, LPrimes[1]^);
              PowerMod(X, Q^, N, PM^);
              if Not PM^.EQ1(1.0) then  {PM <> 1}
                J:= LenPTab  {Abort loop}
              else if J = LenPTab then begin
                M^.SetTo1(-1.0);
              end;
            end
            else begin
              J:= LenPTab;  {Abort loop}
            end;
            J:= J + 1;
          end;
        end; {while}
        if Not PM^.EQ1(1.0) and Not M^.GT1(1.0) then begin
                     {(PM <> 1) and (M < 2)}
          for EI:= 0 to Echo do begin
            Write( OutP[EI]^, 'Large factor ');  N.Writ( OutP[EI]^);
            WriteLn( OutP[EI]^, ' is prime');
          end;
          if ADiagOn then  ADiag('Should exit LargeFactor');
	  Di^.SetTo(N);  Po:= 1;  N.SetTo1(1.0);  Found:= True;
          if Report then  FindAPrime:= False;
        end;
      end;
    end;
  end;
  if M^.GT1(1.0) then begin  {M > 1}
    if KeyHit or ADiagOn then begin
      KeyHit:= False;
      for EI:= 0 to Echo do  WriteLn( OutP[EI]^, 'Large factor is not prime');
    end;
    if Report and FindAPrime then begin
      ExitFactor:= True;
      DelPrimes(LNF,  LPrimes);
      Dispose( NM,    Done);
      Dispose( NP,    Done);
      Dispose( M,     Done);
      Dispose( Q,     Done);
      Dispose( PM,    Done);
      Dispose( Start, Done);
      if ADiagOn then  ADiag('Exit FactorN from LargeFactor');
      Exit;
    end;
  end;
  DelPrimes(LNF,  LPrimes);
  Dispose( NM,    Done);
  Dispose( NP,    Done);
  Dispose( M,     Done);
  Dispose( Q,     Done);
  Dispose( PM,    Done);
  Dispose( Start, Done);
  if ADiagOn then  ADiag('Exit LargeFactor');
end; {LargeFactor}

{--------------------------------------}
procedure SieveMethod; {Part of FactorN}
const LMax = 2000; {Number of iterations of Sieve method per call}
var
  Q       : ^MultiSI;{Quotient}
  R       : ^MultiSI;{Remainder of divide}
  X       : ^MultiSI;{X for search, factor = X - Y}
  Y       : ^MultiSI;{If Y**2 = X**2 - N, then X - Y is the desired factor}
  DP      : ^MultiSI;{Double precision temp}
  Root    : ^MultiSI;{Square root temp}
  Rem     : ^MultiSI;{Remainder of square root}
  Si      : Integer; {Sign result of AbComp procedure}
  KJ      : Integer; {Jth entry of Sieve K table}
  I       : Integer; {Index}
  J       : Integer; {Index}
  L       : Integer; {Count of the number of iterations of Sieve method}
  M       : Integer; {First RS primes}
  MM      : Integer; {(-X) mod M}
  RZero   : Boolean; {True if Remainder of divide is zero}
  AllTrue : Boolean; {All match made in Sieve table are True}
  MATCH   : Boolean; {Match True or False for Sieve table}
  SNF     : Integer; {Number of prime factors found}
  SPrimes : TPrimes; {Prime factors found}
  SPowers : TPowers; {Powers of the prime factors found}
  T       : Double;  {Value of DosClock}

{--------------------------------------}
procedure InitSieve; {Part of SieveMethod}
var
  I   : Integer;  {Index}
  J   : Integer;  {Index}
  L   : Integer;  {Index}
  Dig : Mu1Digit; {Result of one digit Modudo function}
begin
  SieveN^.SetTo(N);
  for I:= 1 to RS do begin
    M:= PTab[I + 1];
    for J:= 0 to M - 1 do
      SqrMod[J]:= (J * J) mod M;
    for J:= 0 to M - 1 do begin
      DP^.Add1(N, -SqrMod[J]);  {DP = N - SqrMod[J]}
      while DP^.S do  DP^.Add1(DP^, M);  {DP = DP + M}
      DP^.Mod1(M, Dig);  {Need -DP mod M}
      MM:= M - Round(Dig);
      if MM = M then  MM:= 0;
      MATCH:= False;
      L:= 0;
      while L < M do begin
        if MM = SqrMod[L] then begin
          MATCH:= True;  L:= M;
        end;
        L:= L + 1;
      end;
      Sieve[I, J]:= MATCH;
      if KeyHit or ADiagOn then begin
        KeyHit:= False;
        for EI:= 0 to Echo do begin
          Write( OutP[EI]^, 'I=', I, ' J=', J, ' MATCH=');
          if MATCH then  WriteLn( OutP[EI]^, 'True')
                   else  WriteLn( OutP[EI]^, 'False');
        end;
      end;
    end;
    if KeyHit or ADiagOn then begin
      KeyHit:= False;
      for EI:= 0 to Echo do  WriteLn( OutP[EI]^);
    end;
  end;
  Root^.SqRtRem(N, RemS^);
  X^.SetTo(Root^);
  if KeyHit or ADiagOn then begin
    KeyHit:= False;
    for EI:= 0 to Echo do begin
      Write( OutP[EI]^, 'Sqrt=');  X^.Writ( OutP[EI]^);
      Write( OutP[EI]^, ' Rem=');  RemS^.WritLn( OutP[EI]^);
    end;
  end;
  if Not RemS^.ZTest then  X^.RAdd1(1.0); {X = X + 1}
  Root2^.Mul1(Root^, 2.0);  {Root2 = Root * 2}
  DelX^.Sub(X^, Root^);      {DelX = X - Root(N)}
  for I:= 1 to RS do begin
    M:= PTab[I + 1];
    X^.Mod1(M, Dig);  {Need -X mod M}
    MM:= M - Round(Dig);
    if MM = M then  MM:= 0;
    if KeyHit or ADiagOn then begin
      KeyHit:= False;
      for EI:= 0 to Echo do  WriteLn( OutP[EI]^, 'KS[', I, ']=', MM);
    end;
    KS[I]:= MM;
  end;
end; {InitSieve}

{--------------------------------------}
procedure ExitSieve; {Part of SieveMethod}
begin
  DelPrimes(SNF,  SPrimes);
  Dispose( Q,    Done);
  Dispose( R,    Done);
  Dispose( X,    Done);
  Dispose( Y,    Done);
  Dispose( DP,   Done);
  Dispose( Root, Done);
  Dispose( Rem,  Done);
  T:= DosClock;
  if ADiagOn then  ADiag('ReportFactor exit SieveMethod');
end; {ExitSieve}

{--------------------------------------}
begin {SieveMethod}
  if NoSieve then Exit;
  if ADiagOn then  ADiag('Enter SieveMethod');
  SNF:= 0;
  New( Q,    Init( Size));
  New( R,    Init( Size));
  New( X,    Init( Size));
  New( Y,    Init( Size));
  New( DP,   Init( Size));
  New( Root, Init( Size));
  New( Rem,  Init( Size));
  for EI:= 0 to Echo do begin
    WriteLn( OutP[EI]^);  Zin^.Writ( OutP[EI]^);
    WriteLn( OutP[EI]^, ' is a hard nut to crack');
  end;
  ExitFactor:= False;
  Found:= False;
  Si:= N.AbComp(SieveN^);
  if Si <> 0 then  InitSieve;
  L:= 0;
  repeat
    AllTrue:= True;
    for I:= 1 to RS do begin
      if Sieve[I, KS[I]] then {NoOp;} else begin
        AllTrue:= False;  I:= RS;
      end;
    end;
    if AllTrue then begin
      L:= L + 1;
      {Root of (X * X - N) needed but X * X may overflow}
      if ADiagOn then begin
        for EI:= 0 to Echo do begin
          Write( OutP[EI]^, 'DelX=X-Root(N)=');  DelX^.WritLn( OutP[EI]^);
        end;
      end;
      Rem^.Add(DelX^, Root2^);  {SqRoot(DelX * (DelX + Root2) - RemS)}
      Rem^.Mul(Rem^, DelX^);
      Rem^.Sub(Rem^, RemS^);
      Root^.SqRtRem(Rem^, Rem^);
      if KeyHit or ADiagOn then begin
        KeyHit:= False;
        if MuAbort then  begin  ExitSieve;  ExitFactor:= True;  Exit;  end;
        Y^.SetTo(Root^);
        DP^.Div1(Root2^, 2.0);  {X = DelX + Root2 div 2}
        X^.Add(DP^, DelX^);
        DP^.Sub(X^, Y^);         {DP = X - Y}
        for EI:= 0 to Echo do begin
          N.ShortWr( OutP[EI]^, 40);  Write( OutP[EI]^, ' / ');
          DP^.ShortWr( OutP[EI]^, 40);  WriteLn( OutP[EI]^, '  SieveMethod');
        end;
      end;
      if ADiagOn then begin
        for EI:= 0 to Echo do begin
          Write( OutP[EI]^, 'Y=Root(X*X - N)=');  Root^.Writ( OutP[EI]^);
          Write( OutP[EI]^, ' Rem=');  Rem^.WritLn( OutP[EI]^);
        end;
      end;
      if Rem^.ZTest then begin
        Y^.SetTo(Root^);
        DP^.Div1(Root2^, 2.0);  {X = DelX + Root2 div 2}
        X^.Add(DP^, DelX^);
        DP^.Sub(X^, Y^);         {DP = X - Y}
        for EI:= 0 to Echo do begin
          DP^.Writ( OutP[EI]^);  Write( OutP[EI]^, ' is a big factor of ');
          N.WritLn( OutP[EI]^);
        end;
        if TwoFactors then begin
          Di^.SetTo(DP^);  Po:= 1;
          DP^.Divi(N, Di^);       {DP = N div Di}
          Si:= Di^.AbComp(DP^);
          if Si = 0 then Po:= Po + 1
          else begin
            N.SetTo(DP^); {This was missing from Apple ][ program}
            ReportFactor;
            if ExitFactor then begin
              ExitSieve;  Exit;
            end;
            Di^.SetTo(DP^);
          end;
          N.SetTo1(1.0);
          ReportFactor;
          if ExitFactor then begin
            ExitSieve;  Exit;
          end;
        end
        else begin
          FactorN(DP^, Start, False, SNF, SPrimes, SPowers);
          Di^.SetTo(SPrimes[1]^);  Po:= SPowers[1];
          for J:= 1 to Po do  N.Divi(N, Di^);  {N = N div Di}
          repeat
            Q^.DivQR(N, Di^, R^);  {Q = N div Di}
            RZero:= R^.ZTest;
            if RZero then begin
              N.SetTo(Q^);  Po:= Po + 1;
            end;
          until Not RZero;
          ReportFactor;
          if ExitFactor then begin
            ExitSieve;  Exit;
          end;
        end;
        Found:= True;  L:= LMax;
      end;
    end;
    {Step X & K'S}
    DelX^.RAdd1(1.0);
    for J:= 1 to RS do begin
      KJ:= KS[J] - 1;
      if KJ < 0 then  KS[J]:= PTab[J + 1] - 1
                else  KS[J]:= KJ;
    end;
  until L >= LMax;
  ExitSieve;
end; {SieveMethod}

{--------------------------------------}
begin {FactorN}
  if ADiagOn then  ADiag('Enter FactorN');
  New( Di, Init( Size));
  if ADiagOn then begin
    for EI:= 0 to Echo do begin
      Write( OutP[EI]^, 'N= ');  N.Writ( OutP[EI]^);
      Write( OutP[EI]^, ' Start= ');  Start.WritLn( OutP[EI]^);
    end;
  end;
  FNF:= 0;
  TwoFactors:= False;
  if Not Start.GT1(Ptab[LenPTab] + 1) then begin  {Start < PTab[LenPTab] + 2}
    SmallFactors;
    if ExitFactor then begin
      Dispose( Di, Done);
      if ADiagOn then  ADiag('SmallFactors exit FactorN');
      Exit;
    end;
    Start.SetTo1(Ptab[LenPTab] + 2);  {Start = PTab[LenPTab] + 2}
    if (FNF > 0) and Report and FindAPrime then begin
      if (FNF = 1) and (FPowers[1] = 1) and N.EQ1(1.0) then begin
        for EI:= 0 to Echo do  WriteLn( OutP[EI]^, 'N is prime');
        FindAPrime:= False;
      end
      else
        for EI:= 0 to Echo do  WriteLn( OutP[EI]^, 'N is not prime');
      Dispose( Di, Done);
      if ADiagOn then  ADiag('Quick exit FactorN');
      Exit;
    end;
    if N.GT1(1.0) then begin
      for EI:= 0 to Echo do begin
        N.Writ( OutP[EI]^);  WriteLn( OutP[EI]^, ' has no small factors');
      end;
    end;
  end;
  SieveN^.Clear;
  Found:= True;
  while N.GT1(1.0) do begin
    repeat
      if Found then begin
        LargeFactor;
        if ExitFactor then begin
          Dispose( Di, Done);
          if ADiagOn then  ADiag('LargeFactor exit FactorN');
          Exit;
        end;
      end;
      if ADiagOn then  ADiag('in FactorN loop');
      if Found then begin
        ReportFactor;
        if ExitFactor then begin
          Dispose( Di, Done);
          if ADiagOn then  ADiag('ReportFactor exit FactorN');
          Exit;
        end;
      end;
      if N.GT1(1.0) then begin
        MoreDiv;
        if ExitFactor then begin
          Dispose( Di, Done);
          if ADiagOn then  ADiag('MoreDiv exit FactorN');
          Exit;
        end;
      end;
    until Not Found or N.EQ1(1.0);
    if N.GT1(1.0) then begin
      SieveMethod;
      if ExitFactor then begin
        Dispose( Di, Done);
        if ADiagOn then  ADiag('SieveMethod exit FactorN');
        Exit;
      end;
    end;
  end;
  Dispose( Di, Done);
  if ADiagOn then  ADiag('Exit FactorN');
end; {FactorN}

{$I GetNum.Inc }

{--------------------------------------}
procedure NextMainNumber; {Main operator loop}
var
  Si     : Integer; {Sign result of AbComp procedure}
  Z      : ^MultiSI;{Factor of ZIn not yet factored}
  Sign   : Integer; {Sign of input number used to flag FindAPrime}
  I      : Integer; {Index}
  J      : Integer; {Index}
  NF     : Integer; {Number of prime factors found in ZIn}
  Primes : TPrimes; {Prime factors found in ZIn}
  Powers : TPowers; {Powers of the prime factors found in ZIn}
begin
  NF:= 0;
  New( Z, Init( Size));
  if MuAbort then  FindAPrime:= False;
  MuErr:= False;  MuAbort:= False;
  if FindAPrime then begin
    Start^.SetTo1(2.0);
    if (FindUp) then ZIn^.RAdd1(+3.0); { to increase }
    ZIn^.RDiv1(2.0);  {ZIn = -((ZIn div 2) * 2 - 1)}
    ZIn^.RMul1(2.0);
    ZIn^.RAdd1(-1.0);
    ZIn^.S:= False;
    for EI:= 0 to Echo do begin
      Write( OutP[EI]^, 'N= ');  ZIn^.WritLn( OutP[EI]^);
    end;
    ZIn^.S:= True;
  end
  else begin
    Str( MemAvail, StIn);
    GetNum( ZIn^,
     'Number to factor: (type Q <Enter> to Quit, MemAvail = '+StIn+')');
  end;
  if Not PTabFilled then begin
    WriteLn( OutP[0]^, 'Wait, filling the ', LenPTab,
      ' entry prime table');
    FillPTab;
    WriteLn(OutP[0]^);
    for I:= LenPTab - 1 to LenPTab do  WriteLn(OutP[0]^, I, '  ', PTab[I]);
    WriteLn(OutP[0]^, 'Sqr Max P = ', SqrMaxP:0:0);
    WriteLn( OutP[0]^);
    PTabFilled:= True;
  end;
  Diag('Start');  Wr2Ln('');
  Sign:= 1;
  FindAPrime:= ZIn^.S and ZIn^.GT1(1.0);
  if ZIn^.S then begin
    ZIn^.S:= False;  Sign:= -1;
  end;
  if ZIn^.N > MuDMax * Size then begin
    for EI:= 0 to Echo do  WriteLn( OutP[EI]^, 'Too big');
    Z^.Clear;
  end
  else  Z^.SetTo(ZIn^);
  if Not Z^.ZTest then begin
    for EI:= 0 to Echo do  WriteLn( OutP[EI]^, Sign);
    if Z^.GT1(1.0) then begin
      TV0:= DosClock;
      if ADiagOn then  ADiag('Starting to factor');
      if ADiagOn then  ADiag('Time Diag procedure');
      FactorN(Z^, Start^, True, NF, Primes, Powers);
      for EI:= 0 to Echo do  WriteLn( OutP[EI]^);
      WriteLn( OutP[EI]^, StIn);
      if Z^.GT1(1.0) then begin
        for EI:= 0 to Echo do begin
          Z^.Writ( OutP[EI]^);
          WriteLn( OutP[EI]^, ' left, not completely factored');
        end;
      end
      else
        for EI:= 0 to Echo do  WriteLn( OutP[EI]^, 'Completely factored');
      for I:= 1 to NF do begin
        for EI:= 0 to Echo do begin
	  Primes[I]^.Writ( OutP[EI]^);
          Write( OutP[EI]^, '^', Powers[I], '  ');
        end;
      end;
      for I:= 1 to NF do begin
        for J:= 1 to Powers[I] do  Z^.RMul(Primes[I]^);  {Z:= Z * Q}
      end;
      for EI:= 0 to Echo do begin
        WriteLn( OutP[EI]^);
        Write( OutP[EI]^, '= ');  Z^.WritLn( OutP[EI]^);
      end;
      Si:= Z^.AbComp(ZIn^);
      if Si <> 0 then {Z <> ZIn}
        for EI:= 0 to Echo do
          WriteLn( OutP[EI]^, 'Error, factors do not check')
    end;
  end;
  DelPrimes(NF,  Primes);
  Dispose( Z, Done);
  Wr2Ln('');  Diag('End of factoring');
end; {NextMainNumber}

{--------------------------------------}
begin {Main program, PrimeFA}
  Init;
  HelpE;
  repeat
    NextMainNumber;
  until False;      {Repeat forever, Halt(0) in GetNum for exit}
end. {PrimeFA}

{End of file PrimeFA.PAS ***************************************************}
